from typing import Tuple

class ProteinMutant:
    # W > Z -> not working
    map_pro = {'ALA': 'A', 'ASH': 'B', 'ASPP': 'B', 'CYS': 'C', 'CYX': 'C', 'CYM': 'C', 'ASP': 'D', 'GLU': 'E',
               'PHE': 'F', 'GLY': 'G', 'HIS': 'H', 'HIE': 'H', 'HID': 'H', 'HSD': 'H', 'HSE': 'H', 'ILE': 'I',
               'GLH': 'J', 'GLPP': 'J', 'LYS': 'K', 'LYN': 'O', 'LEU': 'L', 'MET': 'M', 'ASN': 'N', 'PRO': 'P',
               'GLN': 'Q', 'ARG': 'R', 'SER': 'S', 'THR': 'T', 'VAL': 'V', 'TRP': 'W', 'TYR': 'Y'}
    map_inv = {'A': 'ALA', 'B': 'ASH', 'C': 'CYS', 'D': 'ASP', 'E': 'GLU', 'F': 'PHE', 'G': 'GLY', 'H': 'HIS',
               'I': 'ILE', 'J': 'GLH', 'K': 'LYS', 'L': 'LEU', 'M': 'MET', 'N': 'ASN', 'O': 'LYN', 'P': 'PRO',
               'Q': 'GLN', 'R': 'ARG', 'S': 'SER', 'T': 'THR', 'V': 'VAL', 'W': 'TRP', 'Y': 'TYR'}

    def __init__(self, orig: str, target: str):
        try:
            target_1l = target if len(target) == 1 else ProteinMutant.map_pro[target]
        except KeyError:
            raise RuntimeError("Unrecognized residue specification: {}".format(target))
        try:
            self.name = ProteinMutant.map_pro[orig] + '2' + target_1l
        except KeyError:
            raise RuntimeError("You're trying to mutate residue {} which wasn't found in the database".format(orig))
        self.target_3l = ProteinMutant.map_inv[self.name[-1]]
        orig_gps = self.aminoacids(self.name[0])
        target_gps = self.aminoacids(self.name[-1])
        self.remove_from_orig = []
        self.add_to_target = []
        length = max(len(orig_gps), len(target_gps))
        diverged = False  # if one mismatch is found, side chain will be rebuilt all the way from there
        for i in range(length):
            orig = orig_gps[i] if i < len(orig_gps) else False
            targ = target_gps[i] if i < len(target_gps) else False
            if orig != targ or target_1l in 'IT' or diverged:
                if orig:
                    self.remove_from_orig.append(orig)
                if targ:
                    self.add_to_target.append(targ)
                diverged = True
        if self.name[-1] == 'P':
            self.remove_from_orig.append('PH')
        if self.name[0] == 'P':
            self.add_to_target.append('PH')

    def atoms_to_add(self) -> Tuple[list, list, list, list, list, list]:
        """
        For each atom group, identifies a set of atoms, bonds, anchors etc.
        that have to be included when implementing the mutation
        :return: tuple of lists, the required changes
        """
        # TODO make sure HA1/HA2 are properly assigned in GLY
        atoms = []
        hooks = []
        geo_refs = []
        bond_lengths = []
        topo_bonds = []
        afters = []
        for i in self.add_to_target:
            gatoms = self.groups(i)
            gbonds = self.bonds(i)
            ganchors = self.anchors(i)
            gafters = self.afters(i)
            topo_bonds.extend(self.ring_closing_bonds(i))
            for n, j in enumerate(gatoms):
                atoms.append(j)
                geo_refs.append(ganchors[n])
                hooks.append(gbonds[n][0])
                bond_lengths.append(self.bond_lengths(gbonds[n]))
                afters.append(gafters[n])
        return atoms, hooks, geo_refs, bond_lengths, topo_bonds, afters

    def atoms_to_remove(self) -> list:
        """
        For each atom group, identifies a set of atoms
        that have to be removed when implementing the mutation
        :return:
        """
        atoms = []
        for i in self.remove_from_orig:
            gatoms = self.groups(i)
            for j in gatoms:
                atoms.append(j)
        return atoms

    def check_chiral(self, struct, ca):
        """
        Makes sure the correct HA is removed in case of glycine
        :param struct: gml.Pdb instance
        :param ca: gml.Atom instance
        :return: None
        """
        if 'HA' in self.remove_from_orig:
            if not struct.check_chiral([ca], 'N', 'C', 'HA1', printing=False):
                resnum, resname, chain = ca.resnum, ca.resname, ca.chain
                chn = ' and chain ' + chain if chain.strip() else ''
                ha1 = struct.get_atom(f'name HA1 and resnum {resnum} and resname {resname} {chn}')
                ha2 = struct.get_atom(f'name HA2 and resnum {resnum} and resname {resname} {chn}')
                ha2.atomname = 'HA1'
                ha1.atomname = 'HA2'

    @staticmethod
    def groups(key: str) -> list:
        """
        Defines atoms constituting each atomic group
        (akin e.g. to a methyl or methylene group) in side chains
        :param key: str, name of the group
        :return: list of atoms in the group
        """
        gdict = {'CB': ['CB', 'HB1', 'HB2'],
                 'CG': ['CG', 'HG1', 'HG2'],
                 'CD': ['CD', 'HD1', 'HD2'],
                 'HA': ['HA2'],
                 'HB': ['HB3'],
                 'HD': ['HD3'],
                 'BM': ['CB', 'HB', 'CG2', 'HG21', 'HG22', 'HG23'],
                 'SH': ['SG', 'HG'],
                 'OH': ['OG', 'HG'],
                 'CO': ['CG', 'OD1', 'OD2'],
                 'DO': ['CD', 'OE1', 'OE2'],
                 'CP': ['CG', 'OD1', 'OD2', 'HD2'],
                 'DP': ['CD', 'OE1', 'OE2', 'HE2'],
                 'AM': ['CG', 'OD1', 'ND2', 'HD21', 'HD22'],
                 'AN': ['CD', 'OE1', 'NE2', 'HE21', 'HE22'],
                 'AR': ['CG', 'CD1', 'HD1', 'CD2', 'HD2', 'CE1', 'HE1', 'CE2', 'HE2', 'CZ'],
                 'HG': ['HG3'],
                 'HZ': ['HZ'],
                 'LH': ['CG', 'HG', 'CD2', 'HD21', 'HD22', 'HD23'],
                 'OY': ['OH', 'HH'],
                 'SM': ['SD', 'CE', 'HE1', 'HE2', 'HE3'],
                 'KH': ['CE', 'HE1', 'HE2', 'NZ', 'HZ1', 'HZ2', 'HZ3'],
                 'KN': ['CE', 'HE1', 'HE2', 'NZ', 'HZ1', 'HZ2'],
                 'HR': ['CG', 'ND1', 'CD2', 'HD2', 'CE1', 'NE2', 'HE1', 'HD1'],
                 'GG': ['CG1', 'HG11', 'HG12'],
                 'RH': ['NE', 'HE', 'CZ', 'NH1', 'HH11', 'HH12', 'NH2', 'HH21', 'HH22'],
                 'PD': ['CD', 'HD1', 'HD2'],
                 'PG': ['CG', 'HG1', 'HG2'],
                 'PB': ['CB', 'HB1', 'HB2'],
                 'PH': ['HN'],
                 'WR': ['CG', 'CD1', 'HD1', 'NE1', 'HE1', 'CE2', 'CD2', 'CE3', 'HE3', 'CZ3', 'HZ3', 'CH2', 'HH2', 'CZ2',
                        'HZ2']
                 }
        return gdict[key]

    @staticmethod
    def afters(key: str) -> list:
        """
        For each atomic group, identifies by name atoms immediately
        preceding the ones to be introduced (allowing for variation)
        :param key: str, name of the group
        :return: list of atoms preceding the corresponding atom in self.groups[key] (or tuples if multiple names
        are accepted)
        """
        aftdict = {'CB': [('HA', 'HA2', 'HA1'), 'CB', 'HB1'],
                   'PB': [('HA', 'HA2', 'HA1'), 'CB', 'HB1'],
                   'CG': [('HB2', 'HB'), ('CG', 'CG1'), ('HG1', 'HG11')],
                   'PG': [('HB2', 'HB'), ('CG', 'CG1'), ('HG1', 'HG11')],
                   'CD': [('HG2', 'HG12', 'CD2'), ('CD', 'CD1'), ('HD1', 'HD11')],
                   'PD': [('HG2', 'HG12', 'CD2'), ('CD', 'CD1'), ('HD1', 'HD11')],
                   'HA': [('HA', 'HA1')],
                   'HB': ['HB2'],
                   'HD': [('HD2', 'HD12')],
                   'BM': ['HA', 'CB', 'HB', 'CG2', 'HG21', 'HG22'],
                   'SH': ['HB2', 'SG'],
                   'OH': [('HB2', 'CG2'), ('OG', 'OG1')],
                   'CO': ['HB2', 'CG', 'OD1'],
                   'DO': [('HG2', ), 'CD', 'OE1'],
                   'CP': ['HB2', 'CG', 'OD1', 'OD2'],
                   'DP': [('HG2', ), 'CD', 'OE1', 'OE2'],
                   'AM': ['HB2', 'CG', 'OD1', 'ND2', 'HD21'],
                   'AN': ['HG2', 'CD', 'OE1', 'NE2', 'HE21'],
                   'GG': [('HB2', 'HG23'), 'CG1', 'HG11'],
                   'AR': ['HB2', 'CG', 'CD1', 'HD1', 'CD2', 'HD2', 'CE1', 'HE1', 'CE2', 'HE2'],
                   'HG': [('HG2', 'HG12')],
                   'HZ': ['CZ'], 'LH': ['HB2', 'CG', 'HG', 'CD2', 'HD21', 'HD22'],
                   'OY': ['CZ', 'OH'],
                   'SM': ['HG2', 'SD', 'CE', 'HE1', 'HE2'],
                   'KH': ['HD2', 'CE', 'HE1', 'HE2', 'NZ', 'HZ1', 'HZ2'],
                   'KN': ['HD2', 'CE', 'HE1', 'HE2', 'NZ', 'HZ1'],
                   'HR': ['HB2', 'CG', 'ND1', 'CD2', 'HD2', 'CE1', 'NE2', 'HE1'],
                   'RH': ['HD2', 'NE', 'HE', 'CZ', 'NH1', 'HH11', 'HH12', 'NH2', 'HH21'],
                   'PH': ['N'],
                   'WR': ['HB2', 'CG', 'CD1', 'HD1', 'NE1', 'HE1', 'CE2', 'CD2', 'CE3', 'HE3', 'CZ3', 'HZ3', 'CH2',
                          'HH2', 'CZ2']
                   }
        return aftdict[key]

    @staticmethod
    def bonds(key: str) -> list:
        """
        For each atomic group, identifies chemical bonds that connect atoms
        of that group, as well as the group to the previous one
        :param key: str, name of the group
        :return: list of lists, each containing a pair of atoms to be bonded (or tuples for ambiguous names)
        """
        bonds = {'CB': [['CA', 'CB'], ['CB', 'HB1'], ['CB', 'HB2']],
                 'PB': [['CA', 'CB'], ['CB', 'HB1'], ['CB', 'HB2']],
                 'CG': [['CB', 'CG'], [('CG', 'CG1'), 'HG1'], [('CG', 'CG1'), 'HG2']],
                 'PG': [['CB', 'CG'], [('CG', 'CG1'), 'HG1'], [('CG', 'CG1'), 'HG2']],
                 'GG': [['CB', 'CG1'], ['CG1', 'HG11'], ['CG1', 'HG12']],
                 'CD': [[('CG', 'CG1'), 'CD'], [('CD', 'CD1'), 'HD1'], [('CD', 'CD1'), 'HD2']],
                 'PD': [[('CG', 'CG1'), 'CD'], [('CD', 'CD1'), 'HD1'], [('CD', 'CD1'), 'HD2']],
                 'BM': [['CA', 'CB'], ['CB', 'HB'], ['CB', 'CG2'], ['CG2', 'HG21'], ['CG2', 'HG22'], ['CG2', 'HG23']],
                 'SH': [['CB', 'SG'], ['SG', 'HG']],
                 'OH': [['CB', 'OG'], [('OG', 'OG1'), 'HG']],
                 'CO': [['CB', 'CG'], ['CG', 'OD1'], ['CG', 'OD2']],
                 'DO': [['CG', 'CD'], ['CD', 'OE1'], ['CD', 'OE2']],
                 'CP': [['CB', 'CG'], ['CG', 'OD1'], ['CG', 'OD2'], ['OD2', 'HD2']],
                 'DP': [['CG', 'CD'], ['CD', 'OE1'], ['CD', 'OE2'], ['OE2', 'HE2']],
                 'AM': [['CB', 'CG'], ['CG', 'OD1'], ['CG', 'ND2'], ['ND2', 'HD21'], ['ND2', 'HD22']],
                 'AN': [['CG', 'CD'], ['CD', 'OE1'], ['CD', 'NE2'], ['NE2', 'HE21'], ['NE2', 'HE22']],
                 'AR': [['CB', 'CG'], ['CG', 'CD1'], ['CD1', 'HD1'], ['CG', 'CD2'], ['CD2', 'HD2'], ['CD1', 'CE1'],
                        ['CE1', 'HE1'], ['CD2', 'CE2'], ['CE2', 'HE2'], ['CE2', 'CZ']],
                 'LH': [['CB', 'CG'], ['CG', 'HG'], ['CG', 'CD2'], ['CD2', 'HD21'], ['CD2', 'HD22'], ['CD2', 'HD23']],
                 'OY': [['CZ', 'OH'], ['OH', 'HH']],
                 'SM': [['CG', 'SD'], ['SD', 'CE'], ['CE', 'HE1'], ['CE', 'HE2'], ['CE', 'HE3']],
                 'KH': [['CD', 'CE'], ['CE', 'HE1'], ['CE', 'HE2'], ['CE', 'NZ'], ['NZ', 'HZ1'], ['NZ', 'HZ2'],
                        ['NZ', 'HZ3']],
                 'KN': [['CD', 'CE'], ['CE', 'HE1'], ['CE', 'HE2'], ['CE', 'NZ'], ['NZ', 'HZ1'], ['NZ', 'HZ2']],
                 'PH': [['N', 'H']],
                 'HZ': [['CZ', 'HZ']],
                 'HA': [['CA', 'HA2']],
                 'HB': [['CB', 'HB3']],
                 'HD': [[('CD', 'CD1'), 'HD3']],
                 'HG': [[('CG', 'CG1'), 'HG3']],
                 'HR': [['CB', 'CG'], ['CG', 'ND1'], ['CG', 'CD2'], ['CD2', 'HD2'], ['ND1', 'CE1'], ['CD2', 'NE2'],
                        ['CE1', 'HE1'], ['ND1', 'HD1']],
                 'RH': [['CD', 'NE'], ['NE', 'HE'], ['NE', 'CZ'], ['CZ', 'NH1'], ['NH1', 'HH11'], ['NH1', 'HH12'],
                        ['CZ', 'NH2'], ['NH2', 'HH21'], ['NH2', 'HH22']],
                 'WR': [['CB', 'CG'], ['CG', 'CD1'], ['CD1', 'HD1'], ['CD1', 'NE1'], ['NE1', 'HE1'], ['NE1', 'CE2'],
                        ['CE2', 'CD2'], ['CD2', 'CE3'], ['CE3', 'HE3'], ['CE3', 'CZ3'], ['CZ3', 'HZ3'],
                        ['CZ3', 'CH2'], ['CH2', 'HH2'], ['CH2', 'CZ2'], ['CZ2', 'HZ2']]
                 }
        return bonds[key]

    @staticmethod
    def anchors(key: str) -> list:
        """
        For each atomic group, identifies sequences of atoms that define
        the geometry of the newly built group (through routines defined
        in gml.Pdb._vector)
        :param key: str, name of the group
        :return: list of lists
        """
        # obligatorily same order as in groups
        anchors = {'CB': [['CA', ('HA1', 'HA'), 'C', 'N'], ['N', 'CA'], [('HA1', 'HA'), 'CA']],
                   'PB': [['CA', ('HA1', 'HA'), 'C', 'N'], ['N', ('HA1', 'HA')],
                          [('HA1', 'HA'), 'CB', 'CB', 'CB', 'HB1', 'C', 'C', 'C']],
                   'CG': [['CB', 'CA', ('HB1', 'HB'), ('HB2', 'CG2')], [('HB2', 'CG2'), 'CB'], [('HB1', 'HB'), 'CB']],
                   'PG': [['CA', 'CB', 'CB', 'N', 'N', 'N'], [('HA1', 'HA'), 'CB'], ['C', 'CA']],
                   'GG': [['CB', 'CA', 'HB', 'CG2'], ['CG2', 'CB'], ['HB', 'CB']],
                   'CD': [[('CG', 'CG1'), 'CB', ('HG1', 'HG11', 'HG'), ('HG2', 'HG12', 'CD2')],
                          [('HG2', 'HG12', 'CD2'), ('CG', 'CG1')],
                          [('HG1', 'HG11', 'HG'), ('CG', 'CG1')]],
                   'PD': [['CB', 'N'], ['CA', 'N', 'C', 'CB', 'CA'], ['CD', 'N', 'CG', 'HD1']],
                   'HA': [['CA', 'N', 'C', ('HA1', 'HA')]],
                   'HB': [['CB', 'CA', 'HB1', 'HB2']],
                   'HD': [['CD', ('CG', 'CG1'), 'HD1', 'HD2']],
                   'BM': [['CA', ('HA1', 'HA'), 'C', 'N'], ['C', 'CA'], [('HA1', 'HA'), 'CA'], ['CA', 'CB'],
                          ['CA', 'C'], ['CA', 'N']],
                   'SH': [['CB', 'CA', 'HB1', 'HB2'], ['CA', 'CB']],
                   'OH': [['CB', 'CA', ('HB1', 'HB'), ('HB2', 'CG2')], ['CA', 'CB']],
                   'CO': [['CB', 'CA', 'HB1', 'HB2'], ['CA', 'CB'], ['CG', 'CG', 'CB', 'OD1']],
                   'DO': [['CG', 'CB', 'HG1', 'HG2'], ['CB', 'CG'], ['CD', 'CD', 'CG', 'OE1']],
                   'CP': [['CB', 'CA', 'HB1', 'HB2'], ['CA', 'CB'], ['CG', 'CG', 'CB', 'OD1'], ['CB', 'CG']],
                   'DP': [['CG', 'CB', 'HG1', 'HG2'], ['CB', 'CG'], ['CD', 'CD', 'CG', 'OE1'], ['CG', 'CD']],
                   'AM': [['CB', 'CA', 'HB1', 'HB2'], ['CA', 'CB'], ['CG', 'CB', 'OD1', 'CG'], ['CB', 'CG'],
                          ['OD1', 'CG']],
                   'AN': [['CG', 'CB', 'HG1', 'HG2'], ['CB', 'CG'], ['CD', 'CG', 'OE1', 'CD'], ['CG', 'CD'],
                          ['OE1', 'CD']],
                   'AR': [['CB', 'CA', 'HB1', 'HB2'], ['CB', 'HB1', 'HB1', 'CA', 'CG', 'CG', 'HB1', 'HB1', 'CA', 'CG'],
                          ['CG', 'CD1', 'CB'], ['CB', 'HB2', 'HB2', 'CA', 'CG', 'CG', 'HB2', 'HB2', 'CA', 'CG'],
                          ['CG', 'CD2', 'CB'], ['CB', 'CG'], ['CG', 'CD1'], ['CB', 'CG'], ['CG', 'CD2'], ['CG', 'CD1']],
                   'HG': [['CG', 'CB', ('HG1', 'HG11'), ('HG2', 'HG12')]],
                   'OY': [['CG', 'CZ'], ['CA', 'CB']],
                   'HZ': [['CG', 'CZ']],
                   'PH': [['CA', 'N', 'N', ('CB', 'HA2'), ('CB', 'HA2')]],
                   'LH': [['CB', 'CA', 'HB1', 'HB2'], ['HB2', 'CB'], ['HB1', 'CB'], ['CB', 'CG'], ['CB', 'CA'],
                          ['CB', 'HB2']],
                   'SM': [['CA', 'CB'], ['CB', 'CG'], ['CG', 'SD'], ['CG', 'HG1'], ['CG', 'HG2']],
                   'KH': [['CB', 'CG'], ['CG', 'HG1'], ['CG', 'HG2'], ['CG', 'CD'], ['CD', 'CE'], ['CD', 'HD1'],
                          ['CD', 'HD2']],
                   'KN': [['CB', 'CG'], ['CG', 'HG1'], ['CG', 'HG2'], ['CG', 'CD'], ['CD', 'CE'], ['CD', 'HD1']],
                   'HR': [['CB', 'CA', 'HB1', 'HB2'], ['CB', 'HB1', 'HB1', 'CA', 'CG', 'CG', 'HB1', 'HB1', 'CA', 'CG'],
                          ['CB', 'HB2', 'HB2', 'CA', 'CG', 'CG', 'HB2', 'HB2', 'CA', 'CG'], ['CG', 'CD2', 'CB'],
                          ['CB', 'CG', 'CD2'], ['CB', 'CG', 'ND1'], ['CE1', 'CE1', 'ND1', 'NE2'],
                          ['ND1', 'ND1', 'CG', 'CE1']],
                   'RH': [['CD', 'HD1', 'HD2', 'CG'], ['CD', 'NE', 'CG'], ['CG', 'CD'], ['CD', 'NE'], ['NE', 'CZ'],
                          ['NE', 'HE'], ['HE', 'NE'], ['NE', 'CZ'], ['NE', 'CD']],
                   # ['CG', 'CD1', 'HD1', 'NE1', 'HE1', 'CE2', 'CD2', 'CE3', 'HE3', 'CZ3', 'HZ3', 'CZ2', 'HZ2',
                   # 'CH2', 'HH2']
                   'WR': [['CB', 'CA', 'HB1', 'HB2'],  # CG
                          ['CB', 'CG', 'CA'],  # CD1
                          ['CG', 'CB', 'CD1'],  # HD1
                          ['CD1', 'CG', 'HD1', 'HD1'],  # NE1
                          ['CG', 'CD1'],  # HE1
                          ['CG', 'CB', 'CD1', 'CD1'],  # CE2
                          ['HE1', 'CD1'],  # CD2
                          ['CD1', 'CD2'],  # CE3
                          ['CE2', 'CD2'],  # HE3
                          ['HD1', 'NE1'],  # CZ3
                          ['CE2', 'CZ3'],  # HZ3
                          ['CD2', 'CE2'],  # CH2
                          ['CD2', 'CH2'],  # HH2
                          ['CD2', 'CD1'],  # CZ2
                          ['CE3', 'CZ2']]  # HZ2
                   }
        return anchors[key]

    @staticmethod
    def ring_closing_bonds(key: str) -> list:
        """
        If a residue contains a ring, this fn defines an extra bond
        that closes the ring (might require special treatment)
        :param key: str, name of the group
        :return: list of tuples, each defines a ring-closing bond
        """
        bonds = {'AR': [('CE1', 'CZ')], 'HR': [('CE1', 'NE2')], 'PD': [('CD', 'N')],
                 'WR': [('CZ2', 'CE2'), ('CG', 'CD2')]}
        return bonds[key] if key in bonds.keys() else []

    @staticmethod
    def aminoacids(key: str) -> list:
        """
        Lists groups that make up individual amino acids
        :param key: str, 1-letter code of the amino acid
        :return: list, groups that define the side chain
        """
        aa = {'A': ['CB', 'HB'],
              'B': ['CB', 'CP'],  # protonated ASP
              'C': ['CB', 'SH'],
              'E': ['CB', 'CG', 'DO'],
              'D': ['CB', 'CO'],
              'F': ['CB', 'AR', 'HZ'],
              'G': ['HA'],
              'H': ['CB', 'HR'],
              'I': ['BM', 'GG', 'CD', 'HD'],
              'J': ['CB', 'CG', 'DP'],  # protonated GLU
              'K': ['CB', 'CG', 'CD', 'KH'],
              'L': ['CB', 'LH', 'CD', 'HD'],
              'M': ['CB', 'CG', 'SM'],
              'N': ['CB', 'AM'],
              'O': ['CB', 'CG', 'CD', 'KN'],  # deprotonated LYS
              'Q': ['CB', 'CG', 'AN'],
              'P': ['PB', 'PG', 'PD'],
              'R': ['CB', 'CG', 'CD', 'RH'],
              'S': ['CB', 'OH'],
              'T': ['BM', 'OH'],
              'V': ['BM', 'CG', 'HG'],
              'W': ['CB', 'WR'],
              'Y': ['CB', 'AR', 'OY']}
        return aa[key]

    @staticmethod
    def bond_lengths(bonded_pair: list) -> float:
        """
        Defines bond lengths of pairs of atoms
        :param bonded_pair: list of str, names of the elements that create the bond
        :return: float, length of the bond
        """
        e1 = bonded_pair[0][0] if not isinstance(bonded_pair[0], tuple) else bonded_pair[0][0][0]
        e2 = bonded_pair[1][0]
        if e1 in 'C' and e2 in 'C':
            return 1.5
        elif e1 in 'CO' and e2 in 'CO':
            return 1.4
        elif e1 in 'CN' and e2 in 'CN':
            return 1.45
        elif e1 in 'CH' and e2 in 'CH':
            return 1.1
        elif e1 in 'NOH' and e2 in 'NOH':
            return 1.0
        elif e1 in 'CS' and e2 in 'CS':
            return 1.8
        elif e1 in 'HS' and e2 in 'HS':
            return 1.35
        else:
            raise RuntimeError('Cannot assign bond length to pair: {} {}'.format(e1, e2))



import gromologist as gml


class Entry:
    """
    A generic class representing a single line in the topology.
    In an entry, the actual content and the comments are kept
    in two separate variables.
    """
    def __init__(self, content, subsection):
        self.subsection = subsection
        semicol_index = content.find(';')
        if semicol_index >= 0:
            self.content = content[:semicol_index].strip().split()
            self.comment = ' ' + content[semicol_index:].rstrip()
        else:
            self.content = content.strip().split()
            self.comment = ''
    
    @staticmethod
    def float_fmt(flt: float, fields: int = 11, dpmax: int = 8):
        """
        When a float of unknown precision is read, we do not want
        to clip off significant digits, but neither do we want
        to add too many decimal places. This function calculates
        how many dps we need to keep not to lose precision when
        handling ff params (by default, we clip to 8).
        :param flt: float, the number to be formatted
        :param fields: how many fields do we need overall in the fmt specifier
        :param dpmax: default limit on the number of decimal places
        :return: str, format specifier
        """
        try:
            nf = len(str(flt).split('.')[1])
        except IndexError:
            nf = 3
        if nf > 12:
            flt = float(str(flt).split('.')[0] + '.' + str(flt).split('.')[1][:12])
            nf = len(str(flt).split('.')[1])
        if nf > dpmax:
            dpmax = nf
            fields = dpmax + 3 + len(str(int(flt)))
        else:
            dpmax = nf
        return "{:>" + str(fields) + "." + str(dpmax) + "f}"

    @property
    def is_comment(self) -> bool:
        if not self.content and self.comment:
            return True
        else:
            return False

    @staticmethod
    def infer_type(val) -> type:
        try:
            _ = int(val)
        except ValueError:
            try:
                _ = float(val)
            except ValueError:
                return str
            else:
                return float
        else:
            return int
        
    def __bool__(self) -> bool:
        if not self.content and not self.comment:
            return False
        return True
    
    def __getitem__(self, item):
        return self.content[item]

    def is_header(self) -> bool:
        if len(self.content) == 0:
            return False
        return True if (self.content[0].strip().startswith('[') and self.content[-1].strip().endswith(']')) else False
    
    def __str__(self) -> str:
        """
        Fallback if no explicit formatting is implemented
        :return: str
        """
        return ' '.join(self.content) + ' ' + self.comment + '\n'
    
    
class EntryBonded(Entry):
    """
    This Entry subclass is intended for entries that correspond
    to bonded interaction (bonds, pairs, angles, dihedrals)
    between specific atoms in the topology
    """
    fstr_suff = {('bonds', '1'): (float, float),
                 ('bonds', '2'): (float, float),
                 ('angles', '1'): (float, float),
                 ('angles', '2'): (float, float),
                 ('angles', '5'): (float, float, float, float),
                 ('angles', '10'): (float, float),
                 ('dihedrals', '9'): (float, float, int),
                 ('dihedrals', '4'): (float, float, int),
                 ('dihedrals', '1'): (float, float, int),
                 ('dihedrals', '3'): (float, float, float, float, float, float),
                 ('dihedrals', '2'): (float, float),
                 ('cmap', '1'): (float,),
                 ('position_restraints', '1'): (float, float, float),
                 ('dihedral_restraints', '1'): (float, float, float),
                 ('constraints', '2'): (float,),
                 ('virtual_sites2', '1'): (float,),
                 ('settles', '1'): (float, float)}

    def __init__(self, content, subsection):
        super().__init__(content, subsection)
        if subsection.header == 'exclusions':
            self.atoms_per_entry = len(self.content) - 1
        else:
            self.atoms_per_entry = type(self.subsection).n_atoms[self.subsection.header]
        self.atom_numbers = tuple([int(x) for x in self.content[:self.atoms_per_entry]])
        self.interaction_type = self.content[self.atoms_per_entry]
        try:
            self.params_per_entry = len(EntryBonded.fstr_suff[(subsection.header, str(self.interaction_type))])
        except KeyError:
            self.params_per_entry = 0
            # type assignment should only be performed when asked to, i.e. outside of constructor, with read_types
        self.types_state_a = None
        self.types_state_b = None
        self.atom_names = None
        self.params_state_a = []
        self.params_state_a_entry = []
        self.params_state_b = []
        self.params_state_b_entry = []
        self.fstr_mod = []
        if len(self.content) > self.atoms_per_entry + 1:
            try:
                self.parse_bonded_params(self.content[self.atoms_per_entry + 1:])
            except Exception as e:
                print("While trying to process line {}, subsection {}:".format(content, self.subsection.header))
                raise e
        self.fstring = " ".join("{:>5d}" for _ in range(self.atoms_per_entry)) + " {:>5s}"

    def _fstr_suff(self, query):
        if self.fstr_mod:
            return self.fstr_mod
        else:
            return EntryBonded.fstr_suff[query]

    def explicit_defines(self):
        if self.params_state_a and isinstance(self.params_state_a[0], str):
            try:
                self.params_state_a = self.subsection.section.top.defines[self.params_state_a[0]]
                self.fstr_mod = [self.infer_type(x) for x in self.params_state_a]
            except:
                pass

    def read_types(self):
        atoms_sub = self.subsection.section.get_subsection('atoms')
        atoms_sub.get_dicts()
        num_to_type_a = atoms_sub.num_to_type
        num_to_type_b = atoms_sub.num_to_type_b
        num_to_name = atoms_sub.num_to_name
        self.types_state_a = tuple(num_to_type_a[num] for num in self.atom_numbers)
        types_state_b = tuple(num_to_type_b[num] for num in self.atom_numbers)
        self.types_state_b = types_state_b if types_state_b != self.types_state_a else None
        self.atom_names = tuple(num_to_name[num] for num in self.atom_numbers)

    def parse_bonded_params(self, excess_params):
        try:
            _ = EntryBonded.fstr_suff[(self.subsection.header, self.interaction_type)]
        except KeyError:
            print((self.subsection.header, self.interaction_type))
            raise RuntimeError("Line '{}' contains unrecognized parameters".format(self.content))
        else:
            # if len(excess_params) == 1 and len(types) > 1:
            #     try:
            #         params = self.subsection.section.top.defines[excess_params[0]]
            #     except KeyError:
            #         raise RuntimeError("Cannot process: ", excess_params)
            #     else:
            #         self.params_state_a = [types[n](prm) for n, prm in enumerate(params)]
            try:
                _ = [float(x) for x in excess_params]
            except ValueError:
                # self.fstr_mod = list(EntryBonded.fstr_suff[(self.subsection.header, self.interaction_type)])
                for n in range(len(excess_params)):
                    try:
                        _ = float(excess_params[n])
                    except ValueError:
                        if not excess_params[n] in self.subsection.section.top.defines.keys():
                            if not self.subsection.section.top.ignore_missing_defines:
                                raise RuntimeError(f'undefined parameter {excess_params[n]} was found, try setting a value '
                                                   f'by specifying define={{"{excess_params[n]}": value}} when '
                                                   f'initializing Top, or set ignore_missing_defines=True')
                            else:
                                self.subsection.section.top.defines.update({excess_params[n]: 0})
                self.fstr_mod = [self.infer_type(x) for x in excess_params]
            types = self._fstr_suff((self.subsection.header, self.interaction_type))
            if len(excess_params) == len(types):
                self.params_state_a = [types[n](prm) for n, prm in enumerate(excess_params[:len(types)])]
            elif len(excess_params) == 2 * len(types):
                self.params_state_a = [types[n](prm) for n, prm in enumerate(excess_params[:len(types)])]
                self.params_state_b = [types[n](prm) for n, prm in enumerate(excess_params[len(types):])]
            else:
                raise RuntimeError("Cannot process: ", excess_params)

    def __str__(self):
        fmt_suff = ""
        for params in [self.params_state_a, self.params_state_b]:
            for parm in params:
                if isinstance(parm, int):
                    fmt_suff = fmt_suff + "{:>6d} "
                elif isinstance(parm, float):
                    fmt_suff = ' ' + fmt_suff + self.float_fmt(parm) + ' '
                elif isinstance(parm, str):
                    if len(parm) > 14:
                        fmt_suff = fmt_suff + "{{:>{}s}} ".format(len(parm)+2)
                    else:
                        fmt_suff = fmt_suff + "{:>15s} "
        fstring = self.fstring + fmt_suff
        return fstring.format(*self.atom_numbers, self.interaction_type, *self.params_state_a, *self.params_state_b) \
            + ' ' + self.comment + '\n'

        
class EntryParam(Entry):
    """
    This Entry subclass represents a line containing force field
    parameters, e.g. bondtypes, angletypes, cmaptypes, pairtypes etc.
    that map a set of atom types to a set of FF-specific values
    """
    def __init__(self, content: str, subsection: "gml.Subsection", processed: bool = False, perres=False):
        super().__init__(content, subsection)
        self.atoms_per_entry = type(self.subsection).n_atoms[self.subsection.header]
        self.types = list(self.content[:self.atoms_per_entry])
        if self.subsection.header == 'cmaptypes' and processed:
            if perres:
                self.modifiers = self.content[self.atoms_per_entry + 1:self.atoms_per_entry + 4]
                self.params = [float(x) for x in self.content[self.atoms_per_entry + 4:]]
            else:
                self.modifiers = self.content[self.atoms_per_entry + 1:self.atoms_per_entry + 3]
                self.params = [float(x) for x in self.content[self.atoms_per_entry + 3:]]
            self.interaction_type = self.content[self.atoms_per_entry]
        elif self.subsection.header == 'cmaptypes' and not processed:
            self.modifiers = []
            self.params = self.content[self.atoms_per_entry + 1:]
            self.interaction_type = self.content[self.atoms_per_entry]
        elif self.subsection.header == 'defaults':
            self.modifiers = self.content
            self.params = []
            self.interaction_type = ''
        elif self.subsection.header == 'atomtypes':
            self.modifiers = self.content[self.atoms_per_entry:self.atoms_per_entry + 4]
            self.params = [float(x) for x in self.content[self.atoms_per_entry + 4:]]
            self.interaction_type = ''
        else:
            self.params = [float(x) for x in self.content[self.atoms_per_entry + 1:]]
            self.modifiers = []
            self.interaction_type = self.content[self.atoms_per_entry]
        if self.subsection.header == 'dihedraltypes':
            if any([self.infer_type(x) == float for x in self.types]):
                self.types = self.content[:2]
                self.interaction_type = self.content[2]
                self.params = [float(x) for x in self.content[3:]]
        if self.subsection.header == 'dihedraltypes' and self.interaction_type in ('9', '4', '1'):
            self.params[-1] = int(self.params[-1])
        self.identifier = self.subsection.header + '-' + '-'.join(self.types) + '-' + self.interaction_type
            
    def format(self) -> str:
        fmt = {('bondtypes', '1'): "{:>8s} {:>8s} {:>6s} {:>13.8f} {:>13.2f} ",
               ('angletypes', '5'): "{:>8s} {:>8s} {:>8s} {:>6s} {:>13.6f} {:>13.6f} {:>13.8f} {:>13.2f} ",
               ('angletypes', '1'): "{:>8s} {:>8s} {:>8s} {:>6s} {:>13.8f} {:>13.2f} ",
               ('dihedraltypes', '9'): "{:>8s} {:>8s} {:>8s} {:>8s} {:>6s} {:>13.6f} {:>13.6f} {:>6d} ",
               ('dihedraltypes', '4'): "{:>8s} {:>8s} {:>8s} {:>8s} {:>6s} {:>13.6f} {:>13.6f} {:>6d} ",
               ('dihedraltypes', '3'): "{:>8s} {:>8s} {:>8s} {:>8s} {:>6s} {:>13.6f} {:>13.6f} {:>13.6f} {:>13.6f} "
                                       "{:>13.6f} {:>13.6f} ",
               ('dihedraltypes', '2'): "{:>8s} {:>8s} {:>8s} {:>8s} {:>6s} {:>13.6f} {:>13.6f} ",
               ('dihedraltypes', '1'): "{:>8s} {:>8s} {:>8s} {:>8s} {:>6s} {:>13.6f} {:>13.6f} {:>6d} ",
               ('atomtypes', ''): "{:>6s} {} {:>6s} {:>13s} {:>9s} {:>3s} {:>16.12f} {:>9.5f} ",
               ('pairtypes', '1'): "{:>8s} {:>8s} {:>3s} {:>16.12f} {:>16.12f} ",
               ('nonbond_params', '1'): "{:>8s} {:>8s} {:>3s} {:>20.16f} {:>20.16f} ",
               ('implicit_genborn_params', ''): " {:8s} {:8.4f} {:8.4f} {:8.4f} {:8.4f} {:8.4f} "}
        if (self.subsection.header, self.interaction_type) in fmt.keys():
            return fmt[(self.subsection.header, self.interaction_type)]
        else:
            return ''
    
    def match(self, ext_typelist: list, int_type: str) -> bool:
        """
        Checks if the entry matches an external atomtype list, given the interaction type
        :param ext_typelist: list of str, types to be checked
        :param int_type: str, integer identifier of the interaction type
        :return: bool, whether the entry matches the query
        """
        if not ext_typelist or len(ext_typelist) != len(self.types):
            return False
        if self.interaction_type == int_type:
            if (ext_typelist[0] == self.types[0] or ext_typelist[1] == self.types[1]
                    or ext_typelist[-1] == self.types[-1] or ext_typelist[-2] == self.types[-2]):
                if all(ext_typelist[i] == self.types[i] for i in range(len(self.types)) if self.types[i] != 'X'):
                    return True
            if (ext_typelist[0] == self.types[-1] or ext_typelist[1] == self.types[-2]
                    or ext_typelist[-1] == self.types[0] or ext_typelist[-2] == self.types[1]):
                if all(ext_typelist[i] == self.types[len(self.types)-i-1] for i in range(len(self.types))
                       if self.types[len(self.types)-i-1] != 'X'):
                    return True
        return False
    
    def __repr__(self) -> str:
        if len(self.params) <= 4:
            return "Parameters entry with atomtypes {}, interaction type {} " \
                   "and parameters {}".format(self.types,
                                              self.interaction_type,
                                              ', '.join([str(x) for x in self.params]))
        else:
            return "Parameters entry with atomtypes {}, interaction type {} " \
                   "and parameters {}...".format(self.types,
                                                 self.interaction_type,
                                                 ', '.join([str(x) for x in self.params[:4]]))
        
    def __str__(self) -> str:
        """
        For cmaptypes, we rearrange lines to retrieve the matrix
        format lost during read-in; for other entry types, we
        delegate formatting to Subsection.fmt
        :return:
        """
        if self.subsection.header == 'cmaptypes':
            nf1 = 6 + len(self.modifiers)
            first = ((nf1 * "{} ")[:-1] + "\\\n").format(*self.types, self.interaction_type, *self.modifiers)
            if len(self.modifiers) == 0:
                self.subsection.section.top.print(f"Note: CMAP section {self.types} is missing resolution specifier")
            npar = len(self.params)
            last = '\\\n'.join([((10 * "{} ")[:-1]).format(*self.params[10*n:10*(n+1)]) for n in range(int(npar/10))])
            if 10 * int(npar/10) != npar:
                last = last + '\\\n' + \
                       (((npar-10*int(npar/10)) * "{} ")[:-1]).format(*self.params[10*int(npar/10):]) + '\n'
            return first + last + '\n'
        elif self.format():
            try:
                return self.format().format(*self.types, self.interaction_type, *self.modifiers, *self.params) +\
                       ' ' + self.comment + '\n'
            except:
                print((*self.types, self.interaction_type, *self.modifiers, *self.params))
        else:
            return super().__str__()

        
class EntryAtom(Entry):
    """
    This Entry subclass corresponds to atoms defined in
    the [ atoms ] section of each molecule
    """
    def __init__(self, content, subsection):
        super().__init__(content, subsection)
        try:
            self.num, self.type, self.resid, self.resname, self.atomname, _, self.charge, self.mass = self.content[:8]
        except ValueError:
            self.num, self.type, self.resid, self.resname, self.atomname, _, self.charge = self.content[:7]
            atomtypes = self.subsection.section.top.parameters.get_subsection('atomtypes')
            matching = [atype for atype in atomtypes if isinstance(atype, EntryParam) and atype.types[0] == self.type]
            try:
                self.mass = float(matching[0].modifiers[1])
            except IndexError:
                self.mass = 0
        self.num, self.resid = int(self.num), int(self.resid)
        self.charge, self.mass = float(self.charge), float(self.mass)
        if len(self.content) == 11:
            self.type_b, self.charge_b, self.mass_b = self.content[8], float(self.content[9]), float(self.content[10])
        else:
            self.type_b, self.charge_b, self.mass_b = None, None, None
        self.fstring = "{:>6d} {:>11s} {:>7d}{:>7s}{:>7s}{:>7d}"

    def __eq__(self, other):
        if not isinstance(other, gml.EntryAtom):
            return False
        for attr in ['atomname', 'num', 'resid', 'resname', 'charge', 'mass', 'type', 'type_b', 'charge_b', 'mass_b', 'molname']:
            if self.__getattribute__(attr) != other.__getattribute__(attr):
                return False
        return True

    def __repr__(self):
        if self.type_b is not None:
            extra = f"(A)/{self.type_b}(B)"
        else:
            extra = ''
        return (f"atom {self.atomname}, type {self.type}{extra} in residue {self.resname}-{self.resid} of "
                f"molecule {self.molname}")

    @property
    def molname(self) -> str:
        """
        returns the name of the molecule this atom is part of
        :return: str, name of the molecule
        """
        return self.subsection.section.mol_name

    @property
    def element(self) -> str:
        return [x for x in self.atomname if not x.isdigit()][0]

    @property
    def ish(self, refstate: str = 'A') -> bool:
        """
        Tells if an atom is a hydrogen (useful e.g. for hydrogen mass repartitioning)
        WARNING not very relevant but might make mistakes with helium
        :param refstate: 'A' or 'B', alchemical state to take into account
        :return: bool, whether the atom is a hydrogen
        """
        if refstate == 'A':
            typecheck = self.type[0].upper() == 'H'
        elif refstate == 'B':
            typecheck = self.type_b[0].upper() == 'H'
        else:
            raise RuntimeError("refstate should be 'A' or 'B'")
        return typecheck

    @property
    def sigma(self) -> float:
        """
        Returns the LJ sigma value of an atom (in nm)
        :return: float, the value of sigma
        """
        entry = self._get_atomtype_entry()
        return float(entry.params[0])

    @property
    def epsilon(self) -> float:
        """
        Returns the LJ epsilon value of an atom (in kJ/mol)
        :return: float, the value of epsilon
        """
        entry = self._get_atomtype_entry()
        return float(entry.params[1])

    @property
    def bound_atoms(self) -> list:
        bonds = self.subsection.section.list_bonds(by_num=True, returning=True)
        bonds_involving_self = [b for b in bonds if self.num in b]
        bonds_indices = [i for b in bonds_involving_self for i in b if i != self.num]
        molatoms = self.subsection.section.atoms
        return [molatoms[i-1] for i in bonds_indices]

    def _get_atomtype_entry(self):
        atomtypes = self.subsection.section.top.parameters.get_subsection('atomtypes')
        try:
            return [e for e in atomtypes if isinstance(e, EntryParam) and e.types[0] == self.type][-1]
        except IndexError:
            raise RuntimeError(f"Couldn't find non-bonded parameters for atomtype {self.type}")

    def __str__(self):
        fstring = self.fstring + self.float_fmt(self.charge) + self.float_fmt(self.mass) + '   '
        if self.type_b:
            alch_fstring = "{:>11s}" + self.float_fmt(self.charge_b) + self.float_fmt(self.mass_b)
            fstring += alch_fstring
            return fstring.format(self.num, self.type, self.resid, self.resname, self.atomname, self.num,
                                  self.charge, self.mass, self.type_b, self.charge_b, self.mass_b) + self.comment + '\n'
        else:
            return fstring.format(self.num, self.type, self.resid, self.resname, self.atomname, self.num,
                                  self.charge, self.mass) + ' ' + self.comment + '\n'



from concurrent.futures import ProcessPoolExecutor
import gromologist as gml
import os
from math import ceil
from shutil import copy
# optional?
import numpy as np
from scipy.integrate import simps
from scipy.optimize import fmin
from sklearn.neighbors import KernelDensity
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import LeaveOneOut


class CrooksPool:
    def __init__(self, struct, top, xtc0, xtc1, nruns=100, length=500, gmx=None, maxwarn=1, convergence=False,
                 alias='free', debug=False, offset=0, nmax=None, init_eq=50, temperature=300, stride=20,
                 plumed='', top2='', tmpi=True, mutate='', pmxff='', wt_only=False, mini=False, struct2='',
                 plumed2='', hbond=False, dt=2, weights=None, random=False, **kwargs):
        """
        A pool of workers that will initialize all simulations
        and collect the results afterwards
        :param struct: str, a valid .gro or .pdb structure file for the system
        :param top: str, a .top topology file for the system
        :param xtc0: str, a trajectory capturing the equilibrium ensemble for state A (can also be .dcd or any other
        format accepted by MDTraj)
        :param xtc1: str, a trajectory capturing the equilibrium ensemble for state B
        :param nruns: int, number of A->B runs (overall, twice this many runs will be performed)
        :param length: int, length of a single trajectory (in ps)
        :param gmx: str, Gromacs executable (gmx by default)
        :param maxwarn: int, max number of warnings accepted by grompp
        :param convergence: bool, whether to do convergence analysis
        :param alias: str, a base name for all output files
        :param debug: bool, whether to print detailed convergence info
        :param offset: int, start numbering workers from this number (useful when more jobs are added post-factum)
        :param kwargs: dict of str:str pairs, will be converted into additional .mdp options
        """
        self.struct = struct
        self.struct2 = struct2 if struct2 else self.struct
        self.top = top
        self.top2 = top2 if top2 else self.top
        self.name = alias
        self.bandwidth = 0
        self.debug = debug
        self.convergence = convergence
        self.xtc = {0: xtc0, 1: xtc1}
        self.random = random
        self.nruns = nruns
        self.stride = stride
        self.mini = mini
        self.plumed = plumed
        self.plumed2 = plumed2 if plumed2 else self.plumed
        self.init_length = init_eq
        self.sim_length = length
        self.extra_args = kwargs
        self.lincs = 'all-bonds' if not hbond else 'h-bonds'
        self.maxwarn = maxwarn
        self.offset = offset
        self.temperature = temperature
        self.mutate = mutate
        self.pmxff = pmxff
        self.wt_only = wt_only
        self.dt = dt / 1000
        self.nmax = nmax if nmax is not None else self.nruns
        self.init_nst = self.init_length * int(1 / self.dt)
        self.nst = self.sim_length * int(1 / self.dt)
        if gmx is None:
            found_gmx = gml.find_gmx_dir()
            if found_gmx[1]:
                self.gmx = found_gmx
            else:
                raise RuntimeError("No valid gmx executable found, please specify it as gmx=/path/to/gmx")
        else:
            self.gmx = gmx
        self.gmx = 'gmx' if gmx is None else gmx
        self.tmpi = tmpi
        self.workers = [Crooks(self, n, lam) for n in range(offset, self.nruns + offset) for lam in [0, 1]]
        self.analysis_only = [Crooks(self, n, lam) for n in range(offset) for lam in [0, 1]]
        if weights is None:
            self.weights = [np.ones(self.nruns) / self.nruns, np.ones(self.nruns) / self.nruns]
        else:
            self.weights = [np.loadtxt(weights.replace('X', '0')), np.loadtxt(weights.replace('X', '1'))]
            for w in self.weights:
                w /= np.sum(w)

    def run(self):
        """
        The main routine: extracts initial frames,
        runs grompp and mdrun in parallel
        :return: None
        """
        self.drop_frames()
        if self.mutate:
            print('mutating structures...'.format())
            with ProcessPoolExecutor() as executor:
                executor.map(self.apply_pmx, [(self.mutate, self.wt_only, w.id, w.initlambda, self.pmxff)
                                              for w in self.workers])
        with ProcessPoolExecutor() as executor:
            executor.map(self.run_worker, [(w, 'mini', self.stride, self.temperature, self.init_nst, self.nst,
                                            self.lincs, self.extra_args, self.mini) for w in self.workers])
        self.mdrun_multi('mini')
        with ProcessPoolExecutor() as executor:
            executor.map(self.run_worker, [(w, 'eq', self.stride, self.temperature, self.init_nst, self.nst,
                                            self.lincs, self.extra_args, self.mini) for w in self.workers])
        self.mdrun_multi('eq')
        with ProcessPoolExecutor() as executor:
            executor.map(self.run_worker, [(w, 'prod', self.stride, self.temperature, self.init_nst, self.nst,
                                            self.lincs, self.extra_args, self.mini) for w in self.workers])
        self.mdrun_multi('prod')

    @staticmethod
    def run_worker(data):
        worker, runtype, stride, temperature, init_nst, nst, lincs, extra_args, mini = data
        if runtype == 'mini':
            worker.prep_runs('eq', stride, temperature, init_nst, nst, lincs, extra_args)
            if mini:
                worker.prep_runs(runtype, stride, temperature, init_nst, nst, lincs, extra_args)
                # We can minimize if required
                worker.log('running energy minimization...'.format())
                worker.grompp_me(runtype)
            else:
                worker.cp('run{id}_l{lam}/frame{id}_l{lam}.gro'.format(id=worker.id, lam=worker.initlambda),
                          'mini{id}_l{lam}.gro'.format(id=worker.id, lam=worker.initlambda))
        elif runtype == 'eq':
            # First there's equilibration:
            worker.prep_runs(runtype, stride, temperature, init_nst, nst, lincs, extra_args)
            worker.log('running equilibration...')
            worker.grompp_me(runtype)
        elif runtype == 'prod':
            # The actual alchemistry:
            worker.prep_runs('prod', stride, temperature, init_nst, nst, lincs, extra_args)
            worker.log('running grompp for all systems...')
            worker.grompp_me('prod')
            worker.log('running mdrun for all systems...')

    def analyze(self):
        """
        The analysis routine: performed after the run,
        reads all data files, processes them and
        prints out the results and/or convergence data
        :return:
        """
        print('analyzing results...')
        all_workers = self.analysis_only + self.workers
        for worker in all_workers:
            worker.analyze_me()
        works_0 = np.array([crk.work for crk in all_workers if crk.initlambda == 0]).reshape(-1, 1)
        # TODO sign
        works_1 = np.array([-crk.work for crk in all_workers if crk.initlambda == 1]).reshape(-1, 1)
        np.savetxt('work_0.dat', works_0, fmt='%10.5f')
        np.savetxt('work_1.dat', works_1, fmt='%10.5f')
        min_val = np.min(np.vstack([works_0, works_1]))
        max_val = np.max(np.vstack([works_0, works_1]))
        diff = max_val - min_val
        min_range = min_val - 0.1 * diff
        max_range = max_val + 0.1 * diff
        grid = np.linspace([min_range], [max_range], 200).reshape(-1, 1)
        dens0 = self.get_opt_kde(works_0, grid)
        dens1 = self.get_opt_kde(works_1, grid)
        np.savetxt('prob_0.dat', np.hstack([grid, dens0]), fmt='%10.5f')
        np.savetxt('prob_1.dat', np.hstack([grid, dens1]), fmt='%10.5f')
        result_bar = self.solve_bar(works_0, works_1)
        np.savetxt('normalized_overlap_{}.dat'.format(self.name), self.bhattacharyya(dens0, dens1, grid),
                   fmt='%10.5f')
        if self.convergence:
            self.analyze_convergence(works_0, works_1, grid)
        with open(f'result_{self.name}.dat', 'w') as out:
            out.write(str(result_bar) + '\n')

    @staticmethod
    def bhattacharyya(d0, d1, grid):
        """
        Calculates the Bhattacharyya (similarity) coefficient between distributions d0 and d1
        :param d0: np.array, rho_0, density of A->B work values
        :param d1: np.array, rho_1, density of B->A -work values
        :param grid: np.array, the corresponding X values
        :return: np.array of shape [1,1]
        """
        return simps(np.sqrt(d0.reshape(-1) * d1.reshape(-1)), grid.reshape(-1)).reshape(-1, 1)

    def analyze_convergence(self, w0, w1, grid):
        """
        Performs a series of bootstraps for a number of
        subsample sizes, then plots the resulting means/
        standard deviations to a file
        :param w0: np.array, work values for A->B
        :param w1: np.array, -work values for A->B
        :param grid: np.array, the X values, as in the original plot
        :return: None
        """
        samples = np.linspace(4, self.nruns, 10).astype(int)
        conv = {ns: [] for ns in samples}
        if self.debug:
            try:
                os.mkdir('bootstrap')
            except FileExistsError:
                pass
        for nsamp in samples:
            print('bootstrap for {} samples...'.format(nsamp))
            with ProcessPoolExecutor(max_workers=10) as executor:
                results = np.array(list(executor.map(self.boot, [(w0, w1, nsamp, grid, i)
                                                                 for i in range(10)]))).reshape(-1)
            conv[nsamp] = [np.mean(results), np.std(results)]

    def boot(self, params):
        """
        Performs a single bootstrap resampling,
        parallelized through ProcessPoolExecutor
        :param params: a 5-element tuple of input parameters:
        A->B work, B->A -work, number of subsamples to draw, original X values, random seed
        :return: float, the KDE solution for the intersection
        """
        w0, w1, nsamp, grid, n = params
        np.random.seed(n)
        w0_sampled = np.random.choice(w0.reshape(-1), nsamp, replace=True).reshape(-1, 1)
        w1_sampled = np.random.choice(w1.reshape(-1), nsamp, replace=True).reshape(-1, 1)
        dens0_sampled = self.get_opt_kde(w0_sampled, grid, self.bandwidth)
        dens1_sampled = self.get_opt_kde(w1_sampled, grid, self.bandwidth)
        result = self.solve_bar(w0_sampled, w1_sampled)
        if self.debug:
            np.savetxt(f'bootstrap/conv_{nsamp}_{n}_{result}.dat', np.hstack([grid, dens0_sampled, dens1_sampled]),
                       fmt='%10.5f')
        return result

    @staticmethod
    def solve_kde(d0, d1, grid):
        """
        Finds the intersection point based
        on KDE densities
        :param d0: np.array,
        :param d1: np.array,
        :param grid: np.array,
        :return: float, abscissa of intersection
        """
        diff = d0 - d1
        avg = 0.5 * (np.argmax(d0) + np.argmax(d1))
        solution = [x for x in np.where(diff[1:] * diff[:-1] < 0)[0]]
        arg = solution[np.argmin([np.abs(x - avg) for x in solution])]
        result = round(0.5 * (grid.reshape(-1)[arg] + grid.reshape(-1)[arg]), 3)
        return result

    @staticmethod
    def solve_cgi(w0, w1):
        """
        Finds the intersection point based on the CGI
        estimator (intersection of two Gaussians)
        :param w0: np.array, work values for A->B
        :param w1: np.array, -work values for B->A
        :return: float, abscissa of intersection
        """
        w0, w1 = w0.reshape(-1), w1.reshape(-1)
        m0, s0 = np.mean(w0), np.std(w0)
        m1, s1 = np.mean(w1), np.std(w1)
        x1 = m0 / s0 ** 2 - m1 / s1 ** 2
        den = 1 / s0 ** 2 - 1 / s1 ** 2
        sq = (m0 - m1) ** 2 / (s0 ** 2 * s1 ** 2) + 2 * den * np.log(s1 / s0)
        mid = (m0 + m1) / 2
        r1, r2 = (x1 - np.sqrt(sq)) / den, (x1 + np.sqrt(sq)) / den
        return round(r1, 3) if np.abs(r1 - mid) < np.abs(r2 - mid) else round(r2, 3)

    def solve_bar(self, w0, w1):
        return bar(w0, w1, self.temperature)

    def get_opt_kde(self, y, x, bandwidth=None):
        """
        Performs KDE (kernel density estimation) with a Gaussian
        kernel, and optimizes the bandwidth using the LeaveOneOut
        cross-validation if requested
        :param y: np.array, work values
        :param x: np.array, X values for the density
        :param bandwidth: if None, optimize; if a float, use this value
        :return: np.array, the estimated density
        """
        if bandwidth is None:
            bandwidths = 10 ** np.linspace([-1], [1], 20).reshape(-1)
            grid = GridSearchCV(estimator=KernelDensity(kernel='gaussian'), param_grid={'bandwidth': bandwidths},
                                cv=LeaveOneOut())
            grid.fit(y.reshape(-1, 1))
            kde = KernelDensity(kernel='gaussian', **grid.best_params_).fit(y.reshape(-1, 1))
            self.bandwidth = grid.best_params_['bandwidth']
        else:
            kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth).fit(y.reshape(-1, 1))
        return np.exp(kde.score_samples(x)).reshape(-1, 1)

    def plot_results(self, w0, w1, p0, p1, result_kde, result_cgi, result_bar, result_bar_pmx):
        """
        Plots the two densities, their intersection
        and all individual work values at the bottom;
        also writes the precise solution to a file
        (this is admittedly a bit confusing)
        :param w0: np.array, work values for A->B
        :param w1: np.array, -work values for B->A
        :param p0: np.array, X:Y two columns for the first density
        :param p1: np.array, X:Y two columns for the second density
        :param result_kde: float, KDE result
        :param result_cgi: float, CGI result
        :param result_bar: float, BAR result
        :return: None
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            return
        fig, axes = plt.subplots(2, 1)
        axes[0].plot(*p0.T, c='C0')
        axes[0].plot(*p1.T, c='C1')
        with open('dG_{}.dat'.format(self.name), 'w') as outfile:
            axes[0].axvline(result_kde, ls='--', c='k', lw=3)
            outfile.write("KDE: " + str(result_kde) + '\n')
            outfile.write("CGI: " + str(result_cgi) + '\n')
            outfile.write("BAR: " + str(result_bar) + '\n')
            outfile.write("BAR-PMX: " + str(result_bar_pmx) + '\n')
        axes[0].fill_between(*p0.T, color='C0', alpha=0.25)
        axes[0].fill_between(*p1.T, color='C1', alpha=0.25)
        axes[0].plot(*p1.T, c='C1')
        axes[1].scatter(w0, np.zeros(len(w0)), c='C0')
        axes[1].scatter(w1, np.zeros(len(w1)), c='C1')
        axes[0].set_xlim(np.min(p0[:, 0]), np.max(p0[:, 0]))
        axes[1].set_xlim(np.min(p0[:, 0]), np.max(p0[:, 0]))
        datarange = np.max(p0[:, 0]) - np.min(p0[:, 0])
        axes[1].set_ylim(-datarange / 2, datarange / 2)
        axes[1].get_yaxis().set_visible(False)
        axes[0].get_xaxis().set_visible(False)
        axes[1].set_aspect(0.05)
        axes[1].spines['top'].set_visible(False)
        plt.subplots_adjust(hspace=-0.47)
        plt.savefig('{}.svg'.format(self.name))
        plt.close()

    def drop_frames(self):
        for i in [0, 1]:
            for worker in [w for w in self.workers if w.initlambda == i]:
                worker.mkdir()
                plu = self.plumed if i == 0 else self.plumed2
                if plu:
                    if plu.startswith('/'):
                        plumed = plu
                    else:
                        plumed = os.getcwd() + os.sep + plu
                    worker.cp(plumed, 'plumed.dat')
                    worker.plumed = True
            if not all([os.path.isfile('run{}_l{}/frame{}_l{}.gro'.format(num, i, num, i))
                        for num in range(self.offset, self.offset + self.nruns)]):
                struct = self.struct2 if i == 1 else self.struct
                nfr = gml.frames_count(self.xtc[i], self.gmx)
                if not self.random:
                    frames = np.linspace([0], [nfr - 0.00001], self.nmax).astype(int)[
                             self.offset:self.offset + self.nruns]
                else:
                    frames = np.random.choice(nfr, size=self.nmax, replace=False)
                with open('frames.ndx') as ndxf:
                    ndxf.write(' [ frames ] \n')
                    for m in range(ceil(nfr/15.0)):
                        ndxf.write(' '.join([str(x) for x in frames[15*m:15*(m+1)]]) + '\n')
                gml.gmx_command(self.gmx, 'trjconv', fr='frames.ndx', f=self.xtc[i], s=struct, sep=True, o='frame.gro')
                for fr in range(nfr):
                    os.rename(f'frame{fr}.gro', f'run{fr}_l{i}/frame{fr}_l{i}.gro')
            else:
                print("frames for lambda {} already present, skipping generating new ones".format(i))

    @staticmethod
    def apply_pmx(data):
        mutate, wt_only, num, lam, pmxff = data
        resp = {'A': 'ALA', 'C': 'CYS', 'D': 'ASP', 'E': 'GLU', 'F': 'PHE', 'G': 'GLY', 'H': 'HIS',
                'I': 'ILE', 'K': 'LYS', 'L': 'LEU', 'M': 'MET', 'N': 'ASN', 'P': 'PRO', 'Q': 'GLN',
                'R': 'ARG', 'S': 'SER', 'T': 'THR', 'V': 'VAL', 'W': 'TRP', 'Y': 'TYR'}
        resnums, targets = [], []

        try:
            import pmx
        except ImportError:
            raise RuntimeError("pmx is required for introducing alchemical mutation on-the-fly")

        for mut in mutate.split('/'):
            if mut[1] in '1234567890':
                if lam == 0:
                    target = resp[mut[-1]]
                elif lam == 1 and not wt_only:
                    target = resp[mut[0]]
                else:
                    target = None
                resnum = int(mut[1:-1])
            else:
                if lam == 0:
                    target = mut[-2:]
                elif lam == 1 and not wt_only:
                    target = mut[:2]
                else:
                    target = None
                resnum = int(mut[2:-2])
            resnums.append(resnum)
            targets.append(target)
        model = pmx.Model('run{}_l{}/frame{}_l{}.gro'.format(num, lam, num, lam))
        for resnum, target in zip(resnums, targets):  # TODO OP1, OP2 to O1P, O2P
            if target is not None:
                model = pmx.mutate(model, resnum, target, pmxff)
        model.writeGRO('run{}_l{}/frame{}_l{}.gro'.format(num, lam, num, lam))

    @staticmethod
    def grompp(data):
        crook, runtype = data
        crook.grompp_me(runtype)

    def mdrun_multi(self, runtype):
        tpr = runtype if runtype in ['mini', 'eq'] else 'dyn'
        if self.plumed and runtype == 'eq':
            plu = ' -plumed plumed.dat '
        else:
            plu = ''
        multi = ['run{}_l{}'.format(w.id, w.initlambda) for w in self.workers]
        if not all(['{}.gro'.format(tpr) in os.listdir('run{}_l{}'.format(w.id, w.initlambda)) for w in self.workers]):
            print(gml.gmx_command(f'mpiexec {self.gmx}', 'mdrun', deffnm=tpr, v=plu, cpi=True, multidir=multi,
                                  answer=True))

    @staticmethod
    def read_results(crook):
        crook.analyze_me()


class Crooks:
    def __init__(self, master, num, initlambda):
        """
        A single worker instance
        :param master: a CrooksPool instance
        :param num: ID of the worker
        :param initlambda: initial alchemical lambda (0 or 1)
        """
        self.id = num
        self.master = master
        self.initlambda = initlambda
        self.plumed = False
        self.work = 0

    def mkdir(self):
        """
        Attempts to create a directory if not there yet
        :return: None
        """
        try:
            os.mkdir('run{}_l{}'.format(self.id, self.initlambda))
        except FileExistsError:
            pass

    def cp(self, filename, target_filename):
        copy(filename, 'run{}_l{}/{}'.format(self.id, self.initlambda, target_filename))

    def log(self, command):
        with open('run{}_l{}/crooks.log'.format(self.id, self.initlambda), 'a') as out:
            out.write(command.strip() + '\n')

    def prep_runs(self, runtype, stride, temperature, init_nst, nst, lincs, dt, extra_args):
        """
        Prepares the run by writing customized .mdp files
        :return: None
        """
        print('preparing runs in mode "{}"...'.format(runtype))
        mdp_defaults = {"nstcomm": 100, "nstenergy": 5000, "nstlog": 5000, "nstcalcenergy": 100,
                        "nstxout-compressed": int(1 / dt) * stride, "compressed-x-grps": "System",
                        "compressed-x-precision": 2500, "free-energy": "yes", "sc-alpha": 0.3, "sc-coul": "yes",
                        "sc-sigma": 0.25, "dt": dt, "constraints": lincs, "coulombtype": "PME",
                        "tcoupl": "v-rescale", "ref-t": temperature, "tau-t": 0, "ref-p": 1.0,
                        "rlist": 1.2, "rcoulomb": 1.2, "vdw-type": "Cut-off", "rvdw_switch": 0.8, "rvdw": 1.2,
                        "compressibility": "4.5e-5", "tau-p": 1.0, "tc-grps": "System"}
        mini_defaults = {"integrator": "steep", "free-energy": "yes", "sc-alpha": 0.3, "sc-coul": "yes",
                         "sc-sigma": 0.25, "nsteps": 1000, "emtol": 200, "emstep": 0.001, "nstlist": 10,
                         "pbc": "xyz", "coulombtype": "PME", "vdw-type": "Cut-off"}
        if runtype == 'eq':
            extra_defaults = {"gen-vel": "yes", "gen-temp": temperature, "pcoupl": "Berendsen",
                              "separate-dhdl-file": "no", "nsteps": init_nst, "nstxout": 10000, "nstvout": 10000}
        else:
            extra_defaults = {"gen-vel": "no", "pcoupl": "Parrinello-Rahman", "separate-dhdl-file": "yes",
                              "nsteps": nst, "nstxout": 100000, "nstvout": 100000}
        mdp_defaults.update(extra_defaults)
        mdp_defaults.update(extra_args)
        default = mini_defaults if runtype == 'mini' else mdp_defaults
        pref = 1 if self.initlambda == 0 else -1
        dl = 0 if runtype in ['eq', 'mini'] else pref / nst
        default.update({"init-lambda": self.initlambda, "delta-lambda": dl})
        mdp = '\n'.join(["{} = {}".format(param, value) for param, value in default.items()])
        name = "mini" if runtype == 'mini' else 'md'
        with open('run{}_l{}/{}.mdp'.format(self.id, self.initlambda, name), 'w') as outfile:
            outfile.write(mdp.format(nst=nst, init=self.initlambda, growth=pref / nst))

    def grompp_me(self, runtype):
        """
        Checks for the .tpr file, runs grompp if not present
        :return: None
        """
        tpr = 'eq' if runtype == 'eq' else 'mini' if runtype == 'mini' else 'dyn'
        mdp = 'mini' if runtype == 'mini' else 'md'
        trr = ' ' if runtype in ['mini', 'eq'] else ' -t eq{}_l{}.trr -time {} '.format(self.id, self.initlambda,
                                                                                        self.master.init_length)
        os.chdir('run{}_l{}'.format(self.id, self.initlambda))
        master_top = self.master.top if int(self.initlambda) == 0 else self.master.top2
        frame = 'mini{}_l{}.gro'.format(self.id, self.initlambda) if runtype == 'eq' \
            else 'frame{}_l{}.gro'.format(self.id, self.initlambda)
        if '{}{}_l{}.tpr'.format(tpr, self.id, self.initlambda) not in os.listdir('.'):
            top = master_top if master_top.startswith('/') else '../' + master_top
            print(gml.gmx_command(self.master.gmx, f=f'{mdp}.mdp', p=top, c=frame, answer=True,
                                  o=f'{tpr}{self.id}_l{self.initlambda}.tpr', v=trr, maxwarn=self.master.maxwarn))
            os.remove('mdout.mdp')
        os.chdir('..')

    def analyze_me(self):
        """
        Reads the .xvg file, removes potential duplicates (if the
        job was restarted), and integrates dH/dl to calculate work
        :return: None
        """
        dhdl = np.loadtxt('run{n}_l{l}/dyn.xvg'.format(n=self.id, l=self.initlambda), comments=['#', '@'])
        if not int(dhdl[-1, 0]) == int(self.master.sim_length):
            raise RuntimeError("In file run{}_l{}/dyn.xvg last line reads {}, less than the requested total "
                               "fime of {} ps".format(self.id, self.initlambda, dhdl[-1], self.master.sim_length))
        dhdl_dict = {i: j for i, j in dhdl}
        dhdl = np.array([dhdl_dict[i] for i in sorted(list(dhdl_dict.keys()))])  # avoid duplicates in case of restarts
        endlambda = 1 if self.initlambda == 0 else 0
        self.work = simps(dhdl, np.linspace([self.initlambda], [endlambda], len(dhdl)).reshape(-1))


def bar(forward_works, reverse_works, temperature):
    kb = 0.00831447215  # in kJ units
    forward_works = np.array(forward_works)
    reverse_works = np.array(reverse_works)
    temperature = float(temperature)
    nf = float(len(forward_works))
    nr = float(len(reverse_works))
    beta = 1. / (kb * temperature)
    fact = kb * temperature * np.log(nf / nr)

    def func(x, wf, wr):
        sf = 0
        for v in wf:
            sf += 1 / (1 + np.exp(beta * (fact + v - x)))
        sr = 0
        for v in wr:
            sr += 1 / (1 + np.exp(-beta * (fact + v - x)))
        r = sf - sr
        return r ** 2

    av_a = np.average(forward_works)
    av_b = np.average(reverse_works)
    x0 = (av_a + av_b) / 2.
    dg = fmin(func, x0=x0, args=(forward_works, reverse_works), disp=0)
    return float(dg)
